#!/temptools/bin/bash
#
# *** SEE DOCS FOR HOW TO RUN SCRIPT - ABOVE INTERPRETER PATH IS NOT A TYPO! ***
#
# ---------------------------------------------------------------------------- #
# This file is part of gsbuild
# Copyright (C) 2004 Greg Schafer <gschafer@zip.com.au>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Name:      gsbuild.sh
# Synopsis:  A sample implementation of the "DIY-Linux Reference Build" (x86)
# Features:  * Optional Iterative Comparison Analysis (ICA)
#            * Restartable i.e. uses stamp files to remember where it's up to
# ---------------------------------------------------------------------------- #

# The guts of the temptools phase.
do_temptools() { # temptools_START_MARKER --- DO NOT REMOVE !!!
	# Declare our variables needed for this phase. Export PATH only.
	local PKG_SCRIPTS_DIR="${CURDIR}/scriptlets/${PHASE}"
	local LOGS_DIR="${SCRATCH_DIR}/logs/${PHASE}"
	local STAMP_DIR="${SCRATCH_DIR}/stamp/${PHASE}"
	local TIME_REPORT="${LOGS_DIR}/TIME.REPORT"
	export PATH="${TT_PFX}/bin:/bin:/usr/bin"

	# We need these directories.
	test -d "$LOGS_DIR" || mkdir -p $LOGS_DIR
	test -d "$STAMP_DIR" || mkdir -p $STAMP_DIR

	# Set up the time report.
	time_report start

	# Start building the packages.
	vanilla_build make pass1
	vanilla_build sed pass1
	toolchain_build binutils pass1
	toolchain_build gcc pass1
	vanilla_build linux-libc-headers
	glibc_build
	misc_build adjust-toolchain
	toolchain_build gcc pass2
	toolchain_build binutils pass2
	vanilla_build gawk
	vanilla_build coreutils
	vanilla_build bzip2
	vanilla_build gzip
	vanilla_build diffutils
	vanilla_build findutils
	vanilla_build make pass2
	vanilla_build grep
	vanilla_build sed pass2
	vanilla_build gettext
	vanilla_build ncurses
	vanilla_build patch
	vanilla_build tar
	vanilla_build texinfo
	vanilla_build bash pass2
	vanilla_build m4
	vanilla_build bison
	vanilla_build flex
	vanilla_build util-linux
	vanilla_build perl
	vanilla_build tcl
	vanilla_build expect
	vanilla_build dejagnu
	misc_build remove_cruft

	# Finish up the time report.
	time_report end

	echo -e "\n${GREEN}SUCCESS:${OFF} Build of Temptools Phase Finished!\n"
} # temptools_END_MARKER --- DO NOT REMOVE !!!

# The guts of the chroot phase.
do_chroot() { # chroot_START_MARKER --- DO NOT REMOVE !!!
	# Declare some of our variables needed for this phase.
	local PKG_SCRIPTS_DIR="${CURDIR}/scriptlets/${PHASE}"
	local ITERATIONS=1
	test "$ICA" = Y && ITERATIONS=3

	# Use this variable as the iteration loop counter.
	local ITER=1

	# Ensure the virtual file systems are mounted. If /etc/mtab exists then
	# the real do_mounts was already done and we've likely been restarted.
	test -r /etc/mtab && . ${PKG_SCRIPTS_DIR}/do_mounts

	while [ "$ITER" -le "$ITERATIONS" ]; do

		# Remaining variables needed for this phase. Export PATH only.
		local LOGS_DIR="${SCRATCH_DIR}/logs/${PHASE}-${ITER}"
		local STAMP_DIR="${SCRATCH_DIR}/stamp/${PHASE}-${ITER}"
		local TIME_REPORT="${LOGS_DIR}/TIME.REPORT"

		if [ "$ITER" != 1 ]; then
			export PATH="/bin:/usr/bin:/sbin:/usr/sbin"
		fi

		# We need these directories.
		test -d "$LOGS_DIR" || mkdir -p $LOGS_DIR
		test -d "$STAMP_DIR" || mkdir -p $STAMP_DIR

		# Set up the time report.
		time_report start

		# Start building the packages.
		if [ ! -r "${STAMP_DIR}/../iter${ITER}" ]; then

			if [ "$ITER" = 1 ]; then
				misc_build create_dirs
				misc_build do_mounts
				misc_build create_symlinks
				misc_build create_misc_files
				# When verifying reproducibility, there's little
				# point in reinstalling these next three in
				# subsequent iterations. So put'em here.
				misc_build MAKEDEV
				vanilla_build linux-libc-headers
				vanilla_build man-pages
			fi

			glibc_build

			if [ "$ITER" = 1 ]; then
				misc_build readjust-toolchain
			fi

			toolchain_build binutils
			toolchain_build gcc
			vanilla_build coreutils
			vanilla_build zlib
			vanilla_build debianutils
			vanilla_build iana-etc
			vanilla_build findutils
			vanilla_build gawk
			vanilla_build ncurses
			vanilla_build readline
			vanilla_build vim
			vanilla_build m4
			vanilla_build bison
			vanilla_build less
			vanilla_build groff
			vanilla_build sed
			vanilla_build flex
			vanilla_build gettext
			vanilla_build net-tools
			vanilla_build inetutils
			vanilla_build procps
			vanilla_build perl
			vanilla_build texinfo
			vanilla_build autoconf
			vanilla_build automake
			vanilla_build bash 
			vanilla_build file
			vanilla_build libtool
			vanilla_build bzip2
			vanilla_build diffutils
			#vanilla_build ed
			vanilla_build kbd
			vanilla_build e2fsprogs
			vanilla_build grep
			vanilla_build grub
			vanilla_build gzip
			vanilla_build man
			vanilla_build make
			vanilla_build module-init-tools
			vanilla_build patch
			vanilla_build psmisc
			vanilla_build shadow
			#vanilla_build sysklogd
			vanilla_build sysvinit
			vanilla_build tar
			vanilla_build util-linux

			# Some final cleanups.
			(cd /usr/share/info; rm dir; for f in *; do install-info $f dir 2>/dev/null || :; done)
			rm -rf /tmp/*
			/sbin/ldconfig

			# Do the ICA prep if applicable.
			test "$ICA" = Y && do_ica_prep

			echo -e "\n${GREEN}SUCCESS:${OFF} Build of Chroot Phase Iteration $ITER Finished!\n"

			echo done > ${STAMP_DIR}/../iter${ITER}

		fi

		# We're done. Finish up the time report and bump the counter.
		time_report end
		((ITER += 1))

	done
} # chroot_END_MARKER --- DO NOT REMOVE !!!

# Prepare the scriptlets.
do_prep() {
	check_for_conf_file gsbuild_${BUILD}.prep file
	> $TMP_FILE

	if ! sed -i q $TMP_FILE >/dev/null 2>&1; then
		echo -e "\n${RED}ERROR:${OFF} need sed version 4 or newer! (for the -i option)\n" >&2
		exit 1
	fi

	if [ ! -r "scriptlets/prep-${BUILD}.done" ]; then
		echo -en "\nApplying sed transformations to scriptlets... "
		(. ./gsbuild_${BUILD}.prep) || {
			echo -e "\n${RED}ERROR:${OFF} problem executing gsbuild_${BUILD}.prep!\n" >&2
			exit 1
		}
		echo "done."
		echo done > scriptlets/prep-${BUILD}.done
	else
		echo -e "\n${CYAN}NOTICE:${OFF} scriptlet preparation already done.\n"
		exit 0
	fi
}

# Simple function to check for presence of essential config files/dirs.
check_for_conf_file() {
	test -r "$1" || {
		echo -e "\n\e[1;31mERROR:\e[0m Cannot find requisite \"$1\" $2!\n" >&2
		exit 1
	}
}

# Simple function to show options.
show_options() {
	echo -e "\nOptions Summary:\n"
	echo -e "\tNormal Test Suites:\t`test "$RUN_TESTS" = "Y" && echo Yes || echo No`"
	echo -e "\tOptional Test Suites:\t`test "$RUN_OPTIONAL_TESTS" = "Y" && echo Yes || echo No`"
	echo -e "\tICA Build:\t\t`test "$ICA" = "Y" && echo Yes || echo No`"
}

# Simple function to clean up when exiting via our exit trap.
clean_up () {
	rm -f /tmp/*.$$ >/dev/null 2>&1
	if [ "$PHASE" = chroot ]; then
		umount /proc /dev/{pts,shm} /sys 2>/dev/null
	fi
}

# ---------------------------------------------------------------------------- #
# Main body of script.
# ---------------------------------------------------------------------------- #

# We rely on features found only in bash-2.x. Seeing as we developed and tested
# with 2.05b, just require it as the minimum.
case "$BASH_VERSION" in
2.05[b-z]* | [3-9].*) ;;
*)	echo >&2
	echo "ERROR: Need bash version 2.05b or greater!" >&2
	echo "       Please see docs for info on how to proceed." >&2
	echo >&2
	exit 1 ;;
esac

# Parse the command line args.
PREP=""
DRY_RUN=""
PHASE=""
while [ $# -gt 0 ]; do
	case "$1" in
	--prep)		PREP=1 ;;
	--dry-run)	DRY_RUN=1 ;;
	--chroot)	PHASE=chroot ;;
	*)		echo -e "\nUsage: ${0##*/} [switches]\n" >&2
			echo -e "  --prep\t: prepare the scriptlets" >&2
			echo -e "  --dry-run\t: perform pre-build sanity checks" >&2
			echo -e "  --chroot\t: build target system inside chroot environment\n" >&2
			exit 1 ;;
	esac
	shift
done

# Bomb on simple errors.
# Important - our whole error trapping strategy relies on this. Note - the ERR
# trap is lost inside our while loops as per bash docs.
set -e
trap 'echo -e "\n\e[1;31mERROR:\e[0m Simple error trapped!\n" >&2' ERR

# Clean up on exit. Handle received signals separately.
trap 'clean_up' 0
trap 'exit 2' 1 2 3 15

# Be where we wanna be. This is essential for when relauching ourself upon
# entering the chroot.
cd ${0%/*}

# Initialize some variables.
BUILD=base
CURDIR=`pwd`
: ${PHASE:=temptools}
PKG_LIST=package-${BUILD}.list
TMP_FILE=/tmp/tmp.$$

# Switch off hashing. Set a sane umask.
set +h
umask 022

# Must have `/dev/null' early inside chroot else bad things will happen.
if [ "$PHASE" = chroot ]; then
	test -d /dev || mkdir /dev
	test -c /dev/null || { rm -f /dev/null; mknod /dev/null c 1 3; }
fi

# Check for the 2 main config files. If found, source them, then verify all the
# user supplied environment variables.
check_for_conf_file build.conf file && . ./build.conf
check_for_conf_file common-sh.functions file && . ./common-sh.functions
verify_env_vars "CONFIG_SITE JOBS LC_ALL LDFLAGS TT_PFX TZ"
test "$PHASE" = temptools && verify_env_vars "SYSROOT"
verify_build_conf "SCRATCH_DIR TARBALLS_DIR PATCHES_DIR ICA RUN_TESTS RUN_OPTIONAL_TESTS"
verify_yn_vars "ICA RUN_TESTS RUN_OPTIONAL_TESTS"

# Now check for some other config files/dirs as a basic sanity check.
check_for_conf_file $PKG_LIST file
check_for_conf_file scriptlets/temptools dir
check_for_conf_file scriptlets/chroot dir

# Handle the "--prep" argument.
test "$PREP" && {
	do_prep
	echo -e "\n${GREEN}SUCCESS:${OFF} scriptlet preparation complete!\n"
	exit 0
}

# On some systems 'pwd' returns the "real" cwd instead of what we need e.g:
# "/mnt/sysroot/temptools/src/gsbuild" instead of "/temptools/src/gsbuild".
# Strip off any occurrence of the unwanted prefix.
CURDIR=${CURDIR#${SYSROOT}}

# Now that we have our functions and data, perform some more sanity checks.
verify_dirs_exist "SCRATCH_DIR TARBALLS_DIR PATCHES_DIR TT_PFX"
test "$PHASE" = temptools && verify_dirs_exist "SYSROOT"
verify_dir_writable "SCRATCH_DIR TT_PFX"
verify_dirs "SCRATCH_DIR TARBALLS_DIR PATCHES_DIR"
verify_phase
verify_pkg_list "glibc"
verify_prep
verify_scriptlets temptools glibc
verify_scriptlets chroot glibc
verify_pkg_list_fields
verify_tarballs
verify_patches temptools
verify_patches chroot

# We cannot be superuser in the temptools phase. It's far too dangerous.. unless
# you're a real bloody idiot.
if [ "$PHASE" = temptools -a "`id -u`" = "0" -a "$BLOODYIDIOT" != YES ]; then
	echo -e "\n${RED}ERROR:${OFF} Cannot build Temptools Phase as root!\n" >&2
	exit 1
fi

# If performing an ICA, we don't want to run the test suites. Override.
if [ "$ICA" = Y ]; then
	RUN_TESTS=N
	RUN_OPTIONAL_TESTS=N
fi

# Handle the "--dry-run" argument.
test "$DRY_RUN" && {
	show_options
	echo -e "\n${GREEN}SUCCESS:${OFF} All pre-build sanity checks passed!\n"
	exit 0
}

# Do the business.
if [ "$PHASE" = temptools ]; then
	do_temptools

	CHROOT_CMD="chroot $SYSROOT \
	${TT_PFX}/bin/env -i \
	HOME=/root \
	PATH=/bin:/usr/bin:/sbin:/usr/sbin:${TT_PFX}/bin \
	TERM=${TERM} \
	CONFIG_SITE=/root/config.site \
	JOBS=${JOBS} \
	LC_ALL=${LC_ALL} \
	LDFLAGS=${LDFLAGS} \
	TT_PFX=${TT_PFX} \
	TZ=${TZ} \
	${CURDIR}/${0##*/} --chroot"

	echo -e "${CYAN}NOTICE:${OFF} Please supply root password to enter the chroot environment"
	echo -e "        or press Ctrl-C to exit gracefully.\n"
	test "`id -u`" = "0" -a "$BLOODYIDIOT" != YES && exec "$CHROOT_CMD" || exec su -c "$CHROOT_CMD"
else
	do_chroot

	# Build the ICA reports if applicable.
	if [ "$ICA" = Y ]; then
		if [ ! -r "${SCRATCH_DIR}/stamp/ica" ]; then
			echo -e "${BORDER}\n${CYAN}[ICA] - starting ICA Reports Generation.${OFF}\n"

			do_ica_work 1 2
			do_ica_work 2 3

			echo -e "\n${CYAN}[ICA] - ICA Reports Generation complete.${OFF}\n${BORDER}"
			echo -e "\n${GREEN}SUCCESS:${OFF} Chapter 6 Full ICA Build Finished!\n"
			echo done > ${SCRATCH_DIR}/stamp/ica
		fi
	fi
fi

exit 0
